/*  WAESCipher.cpp -- White-Box Attack Resistant AES cipher implementation
 
    This is part of SecureFW, protection interface for which allows controled usage of secured code.
    See http://www.svenda.com/petr/securefw.html for more information
    Copyright (C) 2004  Petr Svenda

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Please, report any bugs to author <petr@svenda.com>
/**/

#include "WAESCipher.h"
#include "..\Shared\status.h"
#include "SecFunctProvider.h"

CWAESCipher::CWAESCipher() {
    m_pTables = NULL;
    memset(m_IV, 0, BLOCK_LENGTH);
}

int CWAESCipher::SetTables(WBACR_AES_TABLES* pTables) {
    m_pTables = pTables;
    return STAT_OK;
}

//
//  WBACR AES CIPHERING METHODS
//
int CWAESCipher::WBACRAES128_EncryptCBC(BYTE* pbBuffer, DWORD* pcbBufferLen, DWORD* pcbMaxBufferLen, BOOL bUsePadding) {
    int     status = STAT_OK;
    BYTE    j;
    BYTE    bytesToPad = 0;
    DWORD   i;
    DWORD   numBlocks = 0;
    DWORD   offset;

    //
    //  INSERT PADDING
    //
    if (status == STAT_OK) {
        if (bUsePadding) {
            // PKCS#5 PADDING
            status = CSecFunctProvider::InsertCBCPadding(pbBuffer, pcbBufferLen, pcbMaxBufferLen);
            _ASSERT((*pcbBufferLen % BLOCK_LENGTH) == 0);
        }
        else {
            // NO PADDING USED, MUST BE BLOCK_LENGTH MULTIPLY
            if ((*pcbBufferLen % BLOCK_LENGTH) != 0) status = STAT_DATA_INCORRECT_LENGTH;
            _ASSERT((*pcbBufferLen % BLOCK_LENGTH) == 0);
        }
    }


    //
    //  CIPHERING
    //
    if ((status == STAT_OK) && pbBuffer) {
        offset = 0;
        numBlocks = *pcbBufferLen / BLOCK_LENGTH;

        // XOR INIT VECTOR
        for (j = 0; j < BLOCK_LENGTH; j++) pbBuffer[j] ^= m_IV[j];
    
        // ENCRYPT ALL BLOCKS 
        for (i = 0; i < numBlocks; i++) {
            // XOR PREVIOUS ENCIPHERED BLOCK (NOT FOR FIRST BLOCK)
            if (i > 0) for (j = 0; j < BLOCK_LENGTH; j++) pbBuffer[j + offset] ^= pbBuffer[j + offset - BLOCK_LENGTH];

            if ((status = WBACRAES_EncryptOneBlock(pbBuffer + offset)) != STAT_OK) {
                // ENCRYPT FAIL
                break;
            } 

            offset += BLOCK_LENGTH;
        }
    }

    return status;
}

int CWAESCipher::WBACRAES128_DecryptCBC(BYTE* pbBuffer, DWORD* pcbBufferLen, BOOL bUsePadding) {
    int     status = STAT_OK;
    DWORD   i;
    BYTE    j;
    DWORD   numBlocks = *pcbBufferLen / BLOCK_LENGTH;
    DWORD   offset = *pcbBufferLen;
    
    // INPUT BUFFER MUST BE ALLIGNED TO MULTIPLE OF BLOCK_LENGTH
    if (status == STAT_OK) {
        if ((*pcbBufferLen % BLOCK_LENGTH) != 0) {
            status = STAT_DATA_INCORRECT_LENGTH;
        }
    }

    //
    //  DECIPHERING
    //
    if ((status == STAT_OK) && pbBuffer) {
        for (i = numBlocks; i > 0; i--) {
            offset -= BLOCK_LENGTH;

            if ((status = WBACRAES_DecryptOneBlock(pbBuffer + offset)) != STAT_OK) {
                // DECRYPT FAIL
                break;
            }

            // REMOVE CBC XOR
            if (i == 1) {
                // XOR WITH INIT VECTOR
                for (j = 0; j < BLOCK_LENGTH; j++) pbBuffer[j] ^= m_IV[j];
            }
            else {
                // XOR WITH PREVIOUS CIPHERTEXT BLOCK
                for (j = 0; j < BLOCK_LENGTH; j++) pbBuffer[j + offset] ^= pbBuffer[j + offset - BLOCK_LENGTH];
            }
        }
    }

    //
    //  REMOVE PADDING
    //
    if (status == STAT_OK) {
        if (bUsePadding) {
            // CHECK AND REMOVE PKCS#5 PADDING
            status = CSecFunctProvider::RemoveCBCPadding(pbBuffer, pcbBufferLen);
        }
        else {
            // NO PADDING USED
        }  
    }

    return status;
}
/*
//
//  WBACR_AES CIPHERING METHODS WITH RESISTANCE AGAINST VADENAY PKCS#5 PADDING ATTACK   
//
int CWAESCipher::WBACRAES128_EncryptCBC(BYTE* pbBuffer, DWORD* pcbBufferLen, DWORD* pcbMaxBufferLen) {
    int     status = STAT_OK;
    BYTE    j;
    BYTE    bytesToPad = 0;
    DWORD   i;
    DWORD   numBlocks = 0;
    DWORD   offset;

    //
    //  INSERT PADDING
    //
    if (status == STAT_OK) {
        status = CSecFunctProvider::InsertCBCPadding(pbBuffer, pcbBufferLen, pcbMaxBufferLen);
       _ASSERT((*pcbBufferLen % BLOCK_LENGTH) == 0);
    }


    //
    //  CIPHERING
    //
    if ((status == STAT_OK) && pbBuffer) {
        offset = 0;
        numBlocks = *pcbBufferLen / BLOCK_LENGTH;

        // XOR INIT VECTOR
        for (j = 0; j < BLOCK_LENGTH; j++) pbBuffer[j] ^= m_IV[j];
    
        // ENCRYPT ALL BLOCKS BUT LAST
        for (i = 0; i < numBlocks - 1; i++) {
            // XOR PREVIOUS ENCIPHERED BLOCK (NOT FOR FIRST BLOCK)
            if (i > 0) for (j = 0; j < BLOCK_LENGTH; j++) pbBuffer[j + offset] ^= pbBuffer[j + offset - BLOCK_LENGTH];

            if ((status = WBACRAES_EncryptOneBlock(pbBuffer + offset)) != STAT_OK) {
                // ENCRYPT FAIL
                break;
            } 

            offset += BLOCK_LENGTH;
        }

        // STRENGENED LAST BLOCK (DUE TO VADENAY ATTACK) 
        // AS DEFINED AT ftp://ftp.decros.cz/pub/Archiv/Publications/2002/cbc.pdf

        
    }

    return status;
}

int CWAESCipher::WBACRAES128_DecryptCBC(BYTE* pbBuffer, DWORD* pcbBufferLen) {
    int     status = STAT_OK;
    DWORD   i;
    BYTE    j;
    DWORD   numBlocks = *pcbBufferLen / BLOCK_LENGTH;
    DWORD   offset = *pcbBufferLen;
    
    // INPUT BUFFER MUST BE ALLIGNED TO MULTIPLE OF BLOCK_LENGTH
    if (status == STAT_OK) {
        if ((*pcbBufferLen % BLOCK_LENGTH) != 0) {
            status = STAT_DATA_INCORRECT_LENGTH;
        }
    }

    // TODO: TEST FOR CORRECT INITIAL VECTOR

    //
    //  DECIPHERING
    //
    if ((status == STAT_OK) && pbBuffer) {
        for (i = numBlocks; i > 0; i--) {
            offset -= BLOCK_LENGTH;

            if ((status = WBACRAES_DecryptOneBlock(pbBuffer + offset)) != STAT_OK) {
                // DECRYPT FAIL
                break;
            }

            // REMOVE CBC XOR
            if (i == 1) {
                // XOR WITH INIT VECTOR
                for (j = 0; j < BLOCK_LENGTH; j++) pbBuffer[j] ^= m_IV[j];
            }
            else {
                // XOR WITH PREVIOUS CIPHERTEXT BLOCK
                for (j = 0; j < BLOCK_LENGTH; j++) pbBuffer[j + offset] ^= pbBuffer[j + offset - BLOCK_LENGTH];
            }
        }
    }

    //
    //  REMOVE PADDING
    //
    if (status == STAT_OK) {
        status = CSecFunctProvider::RemoveCBCPadding(pbBuffer, pcbBufferLen);
    }

    return status;
}
/**/

int CWAESCipher::WBACRAES_EncryptOneBlock(BYTE plainBlock[BLOCK_LENGTH], BYTE encryptedBlock[BLOCK_LENGTH], int nRounds) {
    //
    // NOTE: INTENDED TO OVERIDE WHEN EMBEDDED WBACR AES TABLES ARE USED   
    //
    int             status = 0;
    int             round;
    int             row;
    int             col;
    int             i;
    int             xorCounter;
    BYTE            XOR_INPUT;
    BYTE            XOR_RES1;
    BYTE            XOR_RES2;
    BYTE            strip[STATE_ROWS];
    AESMATRIX       state;
    AESMATRIX       state2;
    AESMATRIXMC     stateMC;
    DWORD           mc;

    if (!m_pTables) {
        status = -1;
    }

    if (status == STAT_OK) { 
        memset(state, 0, sizeof(AESMATRIX));

        if ((status = CSecFunctProvider::PrepareAESMatrix(plainBlock, BLOCK_LENGTH, state)) == 0) {
            
            //
            // ORDINARY 9 ROUNDS
            //
            for (round = 0; ((round < nRounds) && (round < N_ROUNDS - 1)); round++) {

                //
                //  OBTAIN MC STRIPS (4 BYTES) FOR EACH CELL OF STATE
                //
                for (row = 0; row < STATE_ROWS; row++) {  
                    for (col = 0; col < STATE_COLS; col++) {
                        // OBTAIN MC STRIP FROM POSITION CORRESPONDING TO 
                        //                     |round| |     column               | |row| |    byte value of actual CELL         | |4 == MC STRIP length|
                        memcpy(&mc, m_pTables->ROUND[round] [(col + shifts[row][0]) % 4] [row] [state[row][(col + shifts[row][0]) % 4]], 4);

                        memcpy(stateMC[row][col], &mc, 4);  
                        /* commands above can be also represented by this sequence of commands:
                            AESMATRIXMC stateMC2;
                            stateMC2[row][col][0] =  m_pTables->ROUND[round] [(col + shifts[row][0]) % 4] [row] [state[row][(col + shifts[row][0]) % 4]] [0];
                            stateMC2[row][col][1] =  m_pTables->ROUND[round] [(col + shifts[row][0]) % 4] [row] [state[row][(col + shifts[row][0]) % 4]] [1];
                            stateMC2[row][col][2] =  m_pTables->ROUND[round] [(col + shifts[row][0]) % 4] [row] [state[row][(col + shifts[row][0]) % 4]] [2];
                            stateMC2[row][col][3] =  m_pTables->ROUND[round] [(col + shifts[row][0]) % 4] [row] [state[row][(col + shifts[row][0]) % 4]] [3];
                            _ASSERT(memcmp(stateMC[row][col], stateMC2[row][col], 4) == 0);
                        /**/
                    }
                }

                //
                //  PERFORM XOR OF STRIPS
                //
                for (row = 0; row < STATE_ROWS; row++) {   
                    for (col = 0; col < STATE_COLS; col++) {

                        // PREPARE STRIP FOR GIVEN CELL FROM PRECALC MC STRIPS OF CELLS FROM SAME COLUMN 
                        // DEFINITION OF MIXCOLUMN TRANSFORMATION:
                        // [b0]   [02 03 01 01]   [a0]    b0 = 02*a0 ^ 03*a1 ^ 01*a2 ^ 01*a3
                        // [b1]   [01 02 03 01]   [a1]    b1 = 01*a0 ^ 02*a1 ^ 03*a2 ^ 01*a3
                        // [b2] = [01 01 02 03] * [a2] => b2 = 01*a0 ^ 01*a1 ^ 02*a2 ^ 03*a3
                        // [b3]   [03 01 01 02]   [a3]    b3 = 03*a0 ^ 01*a1 ^ 01*a2 ^ 02*a3
                        // PRECALC MC STRIP CONTAINS {{02*a0, 03*a1, 01*a2, 01*a3}, {01*a2, ..., 01*a0}, ... ,{03*a0, ..., 02*a3}}
                        //        

                        strip[0] = stateMC[0][col][row];
                        strip[1] = stateMC[1][col][row];
                        strip[2] = stateMC[2][col][row];
                        strip[3] = stateMC[3][col][row];
                        //
                        // XOR ADDITIONS OF PREPARED STRIP
                        //
                    
                        // XORING PER 4-BITES OVER STRIP
                        XOR_RES1 = LOWER(GET_LOWER(strip[0]));
                        XOR_RES2 = HIGHER(GET_HIGHER(strip[0]));
                        for (i = 0; i < STATE_ROWS - 1; i++) {
                            // calculating apropriate xor counter for this xoring: 6 xors (== three 8 bits xor => six 4-bits xors) per strip (== row)
                            xorCounter = row * 6 + i * 2;       
                            // XORING LOW 4-BITES 
                            XOR_INPUT = LOWER(GET_LOWER(XOR_RES1)) | HIGHER(GET_LOWER(strip[i + 1]));
                            XOR_RES1 = LOWER(m_pTables->XOR[round][col][xorCounter][XOR_INPUT]);

                            xorCounter = row * 6 + i * 2 + 1;   
                            XOR_INPUT = LOWER(GET_HIGHER(XOR_RES2)) | HIGHER(GET_HIGHER(strip[i + 1]));
                            XOR_RES2 = HIGHER(m_pTables->XOR[round][col][xorCounter][XOR_INPUT]);
                        }

                        state[row][col] = LOWER(GET_LOWER(XOR_RES1));
                        state[row][col] |= HIGHER(GET_HIGHER(XOR_RES2));
                        

                        /* XOR ADDITIONS commands above can be also represented by this sequence of commands:
                            // x*a0 ^ y*a1            
                            xorCounter = row * 6 + (0 == first xor) * 2;   
                            XOR_INPUT = LOWER(GET_LOWER(strip[0])) | HIGHER(GET_LOWER(strip[1]));
                            XOR_RES1 = LOWER(m_pTables->XOR[round][col][xorCounter][XOR_INPUT]);
                            xorCounter = row * 6 + 0 * 2 + 1;   
                            XOR_INPUT = LOWER(GET_HIGHER(strip[0])) | HIGHER(GET_HIGHER(strip[1]));
                            XOR_RES2 = HIGHER(m_pTables->XOR[round][col][xorCounter][XOR_INPUT]);

                            // (x*a0 ^ y*a1) ^ z*a2            
                            xorCounter = row * 6 + (1 == second xor) * 2;       
                            XOR_INPUT = LOWER(GET_LOWER(XOR_RES1)) | HIGHER(GET_LOWER(strip[2]));
                            XOR_RES1 = LOWER(m_pTables->XOR[round][col][xorCounter][XOR_INPUT]);

                            xorCounter = row * 6 + 1 * 2 + 1;   
                            XOR_INPUT = LOWER(GET_HIGHER(XOR_RES2)) | HIGHER(GET_HIGHER(strip[2]));
                            XOR_RES2 = HIGHER(m_pTables->XOR[round][col][xorCounter][XOR_INPUT]);

                            // ((x*a0 ^ y*a1) ^ z*a2) ^ w*a3            
                            xorCounter = row * 6 + (2 == third xor) * 2;       
                            XOR_INPUT = LOWER(GET_LOWER(XOR_RES1)) | HIGHER(GET_LOWER(strip[3]));
                            XOR_RES1 = LOWER(m_pTables->XOR[round][col][xorCounter][XOR_INPUT]);

                            xorCounter = row * 6 + 1 * 2 + 1;   
                            XOR_INPUT = LOWER(GET_HIGHER(XOR_RES2)) | HIGHER(GET_HIGHER(strip[3]));
                            XOR_RES2 = HIGHER(m_pTables->XOR[round][col][xorCounter][XOR_INPUT]);

                            state[row][col] = LOWER(GET_LOWER(XOR_RES1));
                            state[row][col] |= HIGHER(GET_HIGHER(XOR_RES2));
                        /**/

                    }    
                }
            }

            //
            // FINAL ROUND
            //
            if (nRounds == N_ROUNDS) {
                memcpy(state2, state, sizeof(AESMATRIX));
                for (col = 0; col < STATE_COLS; col++) {
                    for (row = 0; row < STATE_ROWS; row++) {
                        state[row][col] = m_pTables->FINALROUND[(col + shifts[row][0]) % 4][row][state2[row][(col + shifts[row][0]) % 4]];
                    }
                }

            }
        }
    }
    
    _ASSERT(status == 0);
    if (status == 0) {
        for (col = 0; col < STATE_COLS; col++) {
            for (row = 0; row < STATE_ROWS; row++) {
                encryptedBlock[col * STATE_COLS + row] = state[row][col];
            }
        }
    }

    return status;    
}

int CWAESCipher::WBACRAES_EncryptOneBlock(BYTE plainBlock[BLOCK_LENGTH], BYTE encryptedBlock[BLOCK_LENGTH]) {
    return WBACRAES_EncryptOneBlock(plainBlock, encryptedBlock, N_ROUNDS);
}

int CWAESCipher::WBACRAES_EncryptOneBlock(BYTE pbBlock[BLOCK_LENGTH]) {
    int     status = STAT_OK;
    BYTE    encrypted[BLOCK_LENGTH];
    
    
    if ((status = WBACRAES_EncryptOneBlock(pbBlock, encrypted)) == STAT_OK) {
        memcpy(pbBlock, encrypted, BLOCK_LENGTH);
    } 
    
    return status;
}
    
int CWAESCipher::WBACRAES_DecryptOneBlock(BYTE encryptedBlock[BLOCK_LENGTH], BYTE plainBlock[BLOCK_LENGTH], int nRounds) {
    //
    // NOTE: INTENDED TO OVERIDE WHEN EMBEDDED WBACR AES TABLES ARE USED   
    //
    int             status = 0;
    int             round;
    int             row;
    int             col;
    int             i;
    int             xorCounter;
    BYTE            XOR_INPUT;
    BYTE            XOR_RES1;
    BYTE            XOR_RES2;
    BYTE            strip[STATE_ROWS];
    AESMATRIX       state;
    AESMATRIX       state2;
    AESMATRIXMC     stateMC;
    DWORD           mc;

    if (!m_pTables) {
        status = -1;
    }

    if (status == STAT_OK) { 
        memset(state, 0, sizeof(AESMATRIX));

        if ((status = CSecFunctProvider::PrepareAESMatrix(encryptedBlock, BLOCK_LENGTH, state)) == 0) {
            

            //
            // INVERSE OF SPECIAL 10. ROUND AND ORDINARY ROUNDS 9 to 2
            //
            for (round = N_ROUNDS; ((round >= nRounds) && (round > 1)); round--) {

                //
                //  OBTAIN MC STRIPS
                //
                for (row = 0; row < STATE_ROWS; row++) {  
                    for (col = 0; col < STATE_COLS; col++) {
                        memcpy(&mc, m_pTables->INVROUND[round-1][(col + shifts[row][1]) % 4][row][state[row][(col + shifts[row][1]) % 4]], 4);
                        memcpy(stateMC[row][col], &mc, 4);
                    }
                }

                //
                //  PERFORM XOR OF STRIPS
                //
                for (row = 0; row < STATE_ROWS; row++) {   
                    for (col = 0; col < STATE_COLS; col++) {

                        // PREPARE STRIP FOR GIVEN CELL FROM PRECALC MC STRIPS OF CELLS FROM SAME COLUMN
                        strip[0] = stateMC[0][col][row];
                        strip[1] = stateMC[1][col][row];
                        strip[2] = stateMC[2][col][row];
                        strip[3] = stateMC[3][col][row];

                        //
                        // XOR ADDITIONS OF PREPARED STRIP
                        //
                    
                        // XORING PER 4-BITES OVER STRIP
                        XOR_RES1 = LOWER(GET_LOWER(strip[0]));
                        XOR_RES2 = HIGHER(GET_HIGHER(strip[0]));
                        for (i = 0; i < STATE_ROWS - 1; i++) {
                            // calculating apropriate xor counter for this xoring: 6 xors per strip (== row)
                            xorCounter = row * 6 + i * 2;       
                            XOR_INPUT = LOWER(GET_LOWER(XOR_RES1)) | HIGHER(GET_LOWER(strip[i + 1]));
                            XOR_RES1 = LOWER(m_pTables->INVXOR[round-1][col][xorCounter][XOR_INPUT]);
                            xorCounter = row * 6 + i * 2 + 1;   
                            XOR_INPUT = LOWER(GET_HIGHER(XOR_RES2)) | HIGHER(GET_HIGHER(strip[i + 1]));
                            XOR_RES2 = HIGHER(m_pTables->INVXOR[round-1][col][xorCounter][XOR_INPUT]);
                        }

                        state[row][col] = LOWER(GET_LOWER(XOR_RES1));
                        state[row][col] |= HIGHER(GET_HIGHER(XOR_RES2));
                    }    
                }
            }

            //
            // INVERSE SPECIAL FIRST ROUND
            //
            if (nRounds == 1) {
                memcpy(state2, state, sizeof(AESMATRIX));
                for (col = 0; col < STATE_COLS; col++) {
                    for (row = 0; row < STATE_ROWS; row++) {
                        state[row][col] = m_pTables->INVFIRSTROUND[(col + shifts[row][1]) % 4][row][state2[row][(col + shifts[row][1]) % 4]];
                    }
                }
            }
        }
    }
    else {
        // NO TABLES
        status = -1;
    }
    
    _ASSERT(status == 0);
    if (status == 0) {
        for (col = 0; col < STATE_COLS; col++) {
            for (row = 0; row < STATE_ROWS; row++) {
                plainBlock[col * STATE_COLS + row] = state[row][col];
            }
        }
    }

    return status;    
}

int CWAESCipher::WBACRAES_DecryptOneBlock(BYTE encryptedBlock[BLOCK_LENGTH], BYTE plainBlock[BLOCK_LENGTH]) {
    return WBACRAES_DecryptOneBlock(encryptedBlock, plainBlock, 1);
}

int CWAESCipher::WBACRAES_DecryptOneBlock(BYTE pbBlock[BLOCK_LENGTH]) {
    int     status = STAT_OK;
    BYTE    decrypted[BLOCK_LENGTH];
    
    
    if ((status = WBACRAES_DecryptOneBlock(pbBlock, decrypted)) == STAT_OK) {
        memcpy(pbBlock, decrypted, BLOCK_LENGTH);
    } 
    
    return status;
}



