/*  WAESCipher_Auth.cpp -- White-Box Attack Resistant AES cipher implementation with embedded tables,
    used for software agent authentization
 
    This is part of SecureFW, protection interface for which allows controled usage of secured code.
    See http://www.svenda.com/petr/securefw.html for more information
    Copyright (C) 2004  Petr Svenda

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Please, report any bugs to author <petr@svenda.com>
/**/

#include "WAESCipher_Auth.h"
#include "..\Shared\status.h"
#include "SecFunctProvider.h"

// EMBEDDED TABLES DEFINITION HEADER FILES
#ifdef EMBEDDED_TABLES_WAESCIPHER_AUTH_TABLES
    #include "Tables\Auth\AESRoundTables_auth.h"
    #include "Tables\Auth\AESFinalTable_auth.h"
    #include "Tables\Auth\AESXorTables_auth.h"
    #include "Tables\Auth\AESInvRoundTables_auth.h"
    #include "Tables\Auth\AESInvXorTables_auth.h"
    #include "Tables\Auth\AESInvFirstTable_auth.h"
#endif

int CWAESCipher_Auth::WBACRAES_EncryptOneBlock(BYTE plainBlock[BLOCK_LENGTH], BYTE encryptedBlock[BLOCK_LENGTH], int nRounds) {
#ifndef EMBEDDED_TABLES_WAESCIPHER_AUTH_TABLES
    //
    // PARENT METHOD CALL
    //
    return CWAESCipher::WBACRAES_EncryptOneBlock(plainBlock, encryptedBlock, nRounds);
#else 
    
    //
    // EMBEDDED TABLES USED, REFER TO CWAESCipher::WBACRAES_EncryptOneBlock FOR MORE COMMENTED CODE
    //

    int             status = STAT_OK;
    int             round;
    int             row;
    int             col;
    int             i;
    int             xorCounter;
    BYTE            XOR_INPUT;
    BYTE            XOR_RES1;
    BYTE            XOR_RES2;
    BYTE            strip[STATE_ROWS];
    AESMATRIX       state;
    AESMATRIX       state2;
    AESMATRIXMC     stateMC;
    DWORD           mc;

    if (status == STAT_OK) { 
        memset(state, 0, sizeof(AESMATRIX));

        if ((status = CSecFunctProvider::PrepareAESMatrix(plainBlock, BLOCK_LENGTH, state)) == 0) {
            
            //
            // ORDINARY 9 ROUNDS
            //
            for (round = 0; ((round < nRounds) && (round < N_ROUNDS - 1)); round++) {

                //
                //  OBTAIN MC STRIPS (4 BYTES) FOR EACH CELL OF STATE
                //
                for (row = 0; row < STATE_ROWS; row++) {  
                    for (col = 0; col < STATE_COLS; col++) {
                        // OBTAIN MC STRIP FROM POSITION CORRESPONDING TO 
                        memcpy(&mc, roundTables_auth[round] [(col + shifts[row][0]) % 4] [row] [state[row][(col + shifts[row][0]) % 4]], 4);
                        memcpy(stateMC[row][col], &mc, 4);  
                    }
                }

                //
                //  PERFORM XOR OF STRIPS
                //
                for (row = 0; row < STATE_ROWS; row++) {   
                    for (col = 0; col < STATE_COLS; col++) {
                        strip[0] = stateMC[0][col][row];
                        strip[1] = stateMC[1][col][row];
                        strip[2] = stateMC[2][col][row];
                        strip[3] = stateMC[3][col][row];
                        //
                        // XOR ADDITIONS OF PREPARED STRIP
                        //
                    
                        // XORING PER 4-BITES OVER STRIP
                        XOR_RES1 = LOWER(GET_LOWER(strip[0]));
                        XOR_RES2 = HIGHER(GET_HIGHER(strip[0]));
                        for (i = 0; i < STATE_ROWS - 1; i++) {
                            // calculating apropriate xor counter for this xoring: 6 xors (== three 8 bits xor => six 4-bits xors) per strip (== row)
                            xorCounter = row * 6 + i * 2;       
                            // XORING LOW 4-BITES 
                            XOR_INPUT = LOWER(GET_LOWER(XOR_RES1)) | HIGHER(GET_LOWER(strip[i + 1]));
                            XOR_RES1 = LOWER(xorTables_auth[round][col][xorCounter][XOR_INPUT]);

                            xorCounter = row * 6 + i * 2 + 1;   
                            XOR_INPUT = LOWER(GET_HIGHER(XOR_RES2)) | HIGHER(GET_HIGHER(strip[i + 1]));
                            XOR_RES2 = HIGHER(xorTables_auth[round][col][xorCounter][XOR_INPUT]);
                        }

                        state[row][col] = LOWER(GET_LOWER(XOR_RES1));
                        state[row][col] |= HIGHER(GET_HIGHER(XOR_RES2));
                    }    
                }
            }

            //
            // FINAL ROUND
            //
            if (nRounds == N_ROUNDS) {
                memcpy(state2, state, sizeof(AESMATRIX));
                for (col = 0; col < STATE_COLS; col++) {
                    for (row = 0; row < STATE_ROWS; row++) {
                        state[row][col] = finalRoundTable_auth[(col + shifts[row][0]) % 4][row][state2[row][(col + shifts[row][0]) % 4]];
                    }
                }

            }
        }
    }
    
    _ASSERT(status == STAT_OK);
    if (status == STAT_OK) {
        for (col = 0; col < STATE_COLS; col++) {
            for (row = 0; row < STATE_ROWS; row++) {
                encryptedBlock[col * STATE_COLS + row] = state[row][col];
            }
        }
    }

    return status;    

#endif // EMBEDDED_TABLES_WAESCIPHER_AUTH_TABLES DEFINED
}

int CWAESCipher_Auth::WBACRAES_DecryptOneBlock(BYTE encryptedBlock[BLOCK_LENGTH], BYTE plainBlock[BLOCK_LENGTH], int nRounds) {
#ifndef EMBEDDED_TABLES_WAESCIPHER_AUTH_TABLES
    //
    // PARENT METHOD CALL
    //
    return CWAESCipher::WBACRAES_DecryptOneBlock(encryptedBlock, plainBlock, nRounds);
#else 
    
    //
    // EMBEDDED TABLES USED, REFER TO CWAESCipher::WBACRAES_DecryptOneBlock FOR MORE COMMENTED CODE
    //

    int             status = 0;
    int             round;
    int             row;
    int             col;
    int             i;
    int             xorCounter;
    BYTE            XOR_INPUT;
    BYTE            XOR_RES1;
    BYTE            XOR_RES2;
    BYTE            strip[STATE_ROWS];
    AESMATRIX       state;
    AESMATRIX       state2;
    AESMATRIXMC     stateMC;
    DWORD           mc;

    if (status == STAT_OK) { 
        memset(state, 0, sizeof(AESMATRIX));

        if ((status = CSecFunctProvider::PrepareAESMatrix(encryptedBlock, BLOCK_LENGTH, state)) == 0) {
            

            //
            // INVERSE OF SPECIAL 10. ROUND AND ORDINARY ROUNDS 9 to 2
            //
            for (round = N_ROUNDS; ((round >= nRounds) && (round > 1)); round--) {

                //
                //  OBTAIN MC STRIPS
                //
                for (row = 0; row < STATE_ROWS; row++) {  
                    for (col = 0; col < STATE_COLS; col++) {
                        memcpy(&mc, invRoundTables_auth[round-1][(col + shifts[row][1]) % 4][row][state[row][(col + shifts[row][1]) % 4]], 4);
                        memcpy(stateMC[row][col], &mc, 4);
                    }
                }

                //
                //  PERFORM XOR OF STRIPS
                //
                for (row = 0; row < STATE_ROWS; row++) {   
                    for (col = 0; col < STATE_COLS; col++) {

                        // PREPARE STRIP FOR GIVEN CELL FROM PRECALC MC STRIPS OF CELLS FROM SAME COLUMN
                        strip[0] = stateMC[0][col][row];
                        strip[1] = stateMC[1][col][row];
                        strip[2] = stateMC[2][col][row];
                        strip[3] = stateMC[3][col][row];

                        //
                        // XOR ADDITIONS OF PREPARED STRIP
                        //
                    
                        // XORING PER 4-BITES OVER STRIP
                        XOR_RES1 = LOWER(GET_LOWER(strip[0]));
                        XOR_RES2 = HIGHER(GET_HIGHER(strip[0]));
                        for (i = 0; i < STATE_ROWS - 1; i++) {
                            // calculating apropriate xor counter for this xoring: 6 xors per strip (== row)
                            xorCounter = row * 6 + i * 2;       
                            XOR_INPUT = LOWER(GET_LOWER(XOR_RES1)) | HIGHER(GET_LOWER(strip[i + 1]));
                            XOR_RES1 = LOWER(invXorTables_auth[round-1][col][xorCounter][XOR_INPUT]);
                            xorCounter = row * 6 + i * 2 + 1;   
                            XOR_INPUT = LOWER(GET_HIGHER(XOR_RES2)) | HIGHER(GET_HIGHER(strip[i + 1]));
                            XOR_RES2 = HIGHER(invXorTables_auth[round-1][col][xorCounter][XOR_INPUT]);
                        }

                        state[row][col] = LOWER(GET_LOWER(XOR_RES1));
                        state[row][col] |= HIGHER(GET_HIGHER(XOR_RES2));
                    }    
                }
            }

            //
            // INVERSE SPECIAL FIRST ROUND
            //
            if (nRounds == 1) {
                memcpy(state2, state, sizeof(AESMATRIX));
                for (col = 0; col < STATE_COLS; col++) {
                    for (row = 0; row < STATE_ROWS; row++) {
                        state[row][col] = invFirstRoundTable_auth[(col + shifts[row][1]) % 4][row][state2[row][(col + shifts[row][1]) % 4]];
                    }
                }
            }
        }
    }
    
    _ASSERT(status == STAT_OK);
    if (status == STAT_OK) {
        for (col = 0; col < STATE_COLS; col++) {
            for (row = 0; row < STATE_ROWS; row++) {
                plainBlock[col * STATE_COLS + row] = state[row][col];
            }
        }
    }

    return status;    

#endif // EMBEDDED_TABLES_WAESCIPHER_AUTH_TABLES DEFINED
}

